//------------------------------------------------------------------------------
// File: RTSP.cpp
//
// Desc: DirectShow sample code - implementation of a renderer that RTSPs
//       the samples it receives into a text file.
//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//------------------------------------------------------------------------------

// Summary
//
// We are a generic renderer that can be attached to any data stream that
// uses IMemInputPin data transport. For each sample we receive we write
// its contents including its properties into a RTSP file. The file we
// will write into is specified when the RTSP filter is created. GraphEdit
// creates a file open dialog automatically when it sees a filter being
// created that supports the IFileSinkFilter interface.
//
//
// Implementation
//
// Pretty straightforward.  We have our own input pin class so that
// we can override Receive, and all that does is to write the properties and
// data into a raw data file (using the Write function). We don't keep
// the file open when we are stopped, so the flags to the open function
// ensure that we open a file if already there; otherwise, we create it.
//
//
// Demonstration instructions
//
// Start GraphEdit, which is available in the SDK DXUtils folder. Drag and drop
// an MPEG, AVI or MOV file into the tool and it will be rendered. Then go to
// the filters in the graph and find the filter (box) titled "Video Renderer"
// This is the filter we will be replacing with the RTSP renderer. Then click
// on the box and hit DELETE. After that go to the Graph menu and select the
// "Insert Filters", from the dialog box find and select the "RTSP Filter".
//
// You will be asked to supply a filename where you would like to have the
// data RTSPed.  (The data we receive in this filter is RTSPed in text form.)
// Then dismiss the dialog. Back in the graph layout, find the output pin of
// the filter that used to be connected to the input of the video renderer
// you just deleted, right click, and do "Render Pin". You should see it being
// connected to the input pin of the RTSP filter you just inserted.
//
// Click Pause and Run and then a little later stop on the GraphEdit frame and
// the data being passed to the renderer will be RTSPed into a file. Stop the
// graph and RTSP the filename that you entered when inserting the filter into
// the graph.  The data supplied to the renderer will be displayed as raw data.
//
//
// Files
//
// RTSP.cpp             Main implementation of the RTSP renderer
// RTSP.def             What APIs the DLL will import and export
// RTSP.h               Class definition of the derived renderer
// RTSP.rc              Version information for the sample DLL
// RTSPuids.h           CLSID for the RTSP filter
// makefile             How to build it...
//
//
// Base classes used
//
// CBaseFilter          Base filter class supporting IMediaFilter
// CRenderedInputPin    An input pin attached to a renderer
// CUnknown             Handle IUnknown for our IFileSinkFilter
// CPosPassThru         Passes seeking interfaces upstream
// CCritSec             Helper class that wraps a critical section
//
//

#include <windows.h>
#include <commdlg.h>
#include <streams.h>
#include <initguid.h>

#include "MyRTSPFilteruids.h"
#include "MyRTSPFilter.h"

//#define DEBUG 

// LiveMedia RTSP 
#include "liveMedia.hh"
#include "GroupsockHelper.hh"
#include "BasicUsageEnvironment.hh"
//#include "MPEG4VideoLiveServerMediaSubsession.h"
#include "MPEG1or2LiveSource.h"

#include <process.h>     // For thread

#include <atlcomcli.h>
#include <strsafe.h>

HANDLE hThread; // Streaming thread.
//extern HANDLE hIOMutex;
extern HANDLE hEvent;
extern char fWatchVariable;

TaskScheduler* scheduler;
UsageEnvironment* env;
struct timeval fPresentationTime; // out

// Uncomment the following if the input file is a MPEG Program Stream
// rather than a MPEG Video Elementary Stream
//#define SOURCE_IS_PROGRAM_STREAM 1

// To stream using "source-specific multicast" (SSM), uncomment the following:
//#define USE_SSM 1
#ifdef USE_SSM
Boolean const isSSM = True;
#else
Boolean const isSSM = False;
#endif

// To set up an internal RTSP server, uncomment the following:
#define IMPLEMENT_RTSP_SERVER 1
// (Note that this RTSP server works for multicast only)

// To stream *only* MPEG "I" frames (e.g., to reduce network bandwidth),
// change the following "False" to "True":
Boolean iFramesOnly = False;

MediaSource* videoSource;
RTPSink* videoSink;

CRTSPInputPin *mpeg1or2LiveSource;
int CRTSPInputPin::FrameIndex = 0;
int CRTSPInputPin::NotReadyFrames = 0;

void play(); // forward

//static char fWatchVariable = 1;

// To make the second and subsequent client for each stream reuse the same
// input stream as the first client (rather than playing the file from the
// start for each client), change the following "False" to "True":
Boolean reuseFirstSource = True;

//// To stream *only* MPEG-1 or 2 video "I" frames
//// (e.g., to reduce network bandwidth),
//// change the following "False" to "True":
//Boolean iFramesOnly = True;

//static void announceStream(RTSPServer* rtspServer, ServerMediaSession* sms,
//			   char const* streamName, char const* inputFileName); // fwd
//
//static void announceStreamLive(RTSPServer* rtspServer, ServerMediaSession* sms,
//			   char const* streamName); // fwd

void InitRTSP(void *);

static void dummyTask(void* /*clientData*/) {
   // Call this again, after a brief delay:
   int uSecsToDelay = 1; // 100 ms
   env->taskScheduler().scheduleDelayedTask(uSecsToDelay, (TaskFunc*)dummyTask, NULL);
}
 //LiveMedia RTSP 

// Setup data

int CRTSPInputPin::MediaSampleCounter = 0;

const AMOVIESETUP_MEDIATYPE sudPinTypes =
{
    &MEDIATYPE_NULL,            // Major type
    &MEDIASUBTYPE_NULL          // Minor type
};

const AMOVIESETUP_PIN sudPins =
{
    L"Input",                   // Pin string name
    FALSE,                      // Is it rendered
    FALSE,                      // Is it an output
    FALSE,                      // Allowed none
    FALSE,                      // Likewise many
    &CLSID_NULL,                // Connects to filter
    L"Output",                  // Connects to pin
    1,                          // Number of types
    &sudPinTypes                // Pin information
};

const AMOVIESETUP_FILTER sudRTSP =
{
    &CLSID_MYRTSPFILTER,                // Filter CLSID
    L"aMyRTSP Filter",                    // String name
    MERIT_DO_NOT_USE,           // Filter merit
    1,                          // Number pins
    &sudPins                    // Pin details
};


//
//  Object creation stuff
//
CFactoryTemplate g_Templates[]= {
    L"aMyRTSP Filter", &CLSID_MYRTSPFILTER, CRTSP::CreateInstance, NULL, &sudRTSP
};
int g_cTemplates = 1;


// Constructor

CRTSPFilter::CRTSPFilter(CRTSP *pRTSP,
                         LPUNKNOWN pUnk,
                         CCritSec *pLock,
                         HRESULT *phr) :
    CBaseFilter(NAME("CRTSPFilter"), pUnk, pLock, CLSID_MYRTSPFILTER),
    m_pRTSP(pRTSP)
{
}


//
// GetPin
//
CBasePin * CRTSPFilter::GetPin(int n)
{
    if (n == 0) {
        return m_pRTSP->m_pPin;
    } else {
        return NULL;
    }
}


//
// GetPinCount
//
int CRTSPFilter::GetPinCount()
{
    return 1;
}


//
// Stop
//
// Overriden to close the RTSP file
//
STDMETHODIMP CRTSPFilter::Stop()
{
    CAutoLock cObjectLock(m_pLock);

    if (m_pRTSP)
        m_pRTSP->CloseFile();
    
    return CBaseFilter::Stop();
}


//
// Pause
//
// Overriden to open the RTSP file
//
STDMETHODIMP CRTSPFilter::Pause()
{
    CAutoLock cObjectLock(m_pLock);

    if (m_pRTSP)
    {
        // GraphEdit calls Pause() before calling Stop() for this filter.
        // If we have encountered a write error (such as disk full),
        // then stopping the graph could cause our log to be deleted
        // (because the current log file handle would be invalid).
        // 
        // To preserve the log, don't open/create the log file on pause
        // if we have previously encountered an error.  The write error
        // flag gets cleared when setting a new log file name or
        // when restarting the graph with Run().
        if (!m_pRTSP->m_fWriteError)
        {
            m_pRTSP->OpenFile();
        }
    }

    return CBaseFilter::Pause();
}


//
// Run
//
// Overriden to open the RTSP file
//
STDMETHODIMP CRTSPFilter::Run(REFERENCE_TIME tStart)
{
    CAutoLock cObjectLock(m_pLock);

    // Clear the global 'write error' flag that would be set
    // if we had encountered a problem writing the previous RTSP file.
    // (eg. running out of disk space).
    //
	// Since we are restarting the graph, a new file will be created.
    m_pRTSP->m_fWriteError = FALSE;

    if (m_pRTSP)
        m_pRTSP->OpenFile();

    return CBaseFilter::Run(tStart);
}


CRTSPInputPin::CRTSPInputPin(UsageEnvironment& env,
				  CRTSP *pRTSP,
                  LPUNKNOWN pUnk,
                  CBaseFilter *pFilter,
                  CCritSec *pLock,
                  CCritSec *pReceiveLock,
				  HRESULT *phr):
    FramedSource(env),
	CRenderedInputPin(NAME("CRTSPInputPin"),
                  pFilter,                   // Filter
                  pLock,                     // Locking
                  phr,                       // Return code
                  L"Input"),                 // Pin name
    m_pReceiveLock(pReceiveLock),
    m_pRTSP(pRTSP),
    m_tLast(0)
{
}

CRTSPInputPin* CRTSPInputPin::createNew(UsageEnvironment& env,
										CRTSP *pRTSP,
										LPUNKNOWN pUnk,
										CBaseFilter *pFilter,
										CCritSec *pLock,
										CCritSec *pReceiveLock,
										HRESULT *phr)
{
	return new CRTSPInputPin(env,pRTSP,pUnk,pFilter,pLock,pReceiveLock,phr);
}

CRTSPInputPin::~CRTSPInputPin() {
}

void CRTSPInputPin::doGetNextFrame()
{
	if (!isCurrentlyAwaitingData()) 
	{	
		printf("\n!!!NotReadyFrames#=%d size=%d",NotReadyFrames++);
		return; // we're not ready for the data yet
	}
		  
	//if (fWatchVariable == 0)        //  if (/*no data is available on your queue*/) return;
	//  return;
	//
	
	if (pbGReadBuffer == NULL) 
	{
		return;
	}

    if (lGActualDataLength < fMaxSize)
	{
		fFrameSize = lGActualDataLength;
	    fNumTruncatedBytes = 0;
	}
	else
	{	
	  fFrameSize = fMaxSize;
	  fNumTruncatedBytes = lGActualDataLength - fMaxSize;
	}
	
    memcpy(fTo,pbGReadBuffer,fFrameSize);
	printf("\ndeliverFrame#=%d size=%d",FrameIndex++,lGActualDataLength);
	for (int i=0;i<7;i++)	printf("=%dx%x,",i,pbGReadBuffer[i]);		

	gettimeofday(&fPresentationTime,NULL);
	printf("\nDel: - sec=%d,msec=%d",fPresentationTime.tv_sec,fPresentationTime.tv_usec);
	//fDurationInMicroseconds = 1000000 / 25; // 25 fps 

    nextTask() = envir().taskScheduler().scheduleDelayedTask(0,(TaskFunc*)FramedSource::afterGetting, this);	
}
//void CRTSPInputPin::deliverFrame()
//{ 
//	if (!isCurrentlyAwaitingData()) 
//	{	
//		printf("\n!!!NotReadyFrames#=%d size=%d",NotReadyFrames++);
//		return;
//	}
////		  return; // we're not ready for the data yet
//	//if (fWatchVariable == 0)        //  if (/*no data is available on your queue*/) return;
//	//	  return;
//	
//	if (pbGReadBuffer == NULL) 
//	{
//		return;
//	}
////#ifdef DEBUG
//	printf("\ndeliverFrame#=%d size=%d",FrameIndex++,lGActualDataLength);
//	for (int i=0;i<7;i++)	printf("=%dx%x,",i,pbGReadBuffer[i]);	
////#endif
//	
//	if (lGActualDataLength < fMaxSize)
//	{
//		fFrameSize = lGActualDataLength;
//	    fNumTruncatedBytes = 0;
//	}
//	else
//	{	
//	  fFrameSize = fMaxSize;
//	  fNumTruncatedBytes = lGActualDataLength - fMaxSize;
//	}
//    memcpy(fTo,pbGReadBuffer,fFrameSize);
//	
//	gettimeofday(&fPresentationTime,NULL);
//	printf("\nDel: - sec=%d,msec=%d",fPresentationTime.tv_sec,fPresentationTime.tv_usec);
//	//fDurationInMicroseconds = 1000000 / 25; // 25 fps 
//	FramedSource::afterGetting(this);
//
//}

//
// CheckMediaType
//
// Check if the pin can support this specific proposed type and format
//
HRESULT CRTSPInputPin::CheckMediaType(const CMediaType *)
{
    return S_OK;
}


//
// BreakConnect
//
// Break a connection
//
HRESULT CRTSPInputPin::BreakConnect()
{
    if (m_pRTSP->m_pPosition != NULL) {
        m_pRTSP->m_pPosition->ForceRefresh();
    }

    return CRenderedInputPin::BreakConnect();
}


//
// ReceiveCanBlock
//
// We don't hold up source threads on Receive
//
STDMETHODIMP CRTSPInputPin::ReceiveCanBlock()
{
    return S_FALSE;
}


//
// Receive the IMediaSample and store in a global buffer
//
STDMETHODIMP CRTSPInputPin::Receive(IMediaSample *pSample)
{
    CheckPointer(pSample,E_POINTER);
    CAutoLock lock(m_pReceiveLock);

    PBYTE pbData;
	if (hThread == NULL)
	{
		hThread = (HANDLE)_beginthread(InitRTSP, 0,hEvent);
		
	}
    // Has the filter been stopped yet?
    if (m_pRTSP->m_hFile == INVALID_HANDLE_VALUE) {
        return NOERROR;
    }

    REFERENCE_TIME tStart, tStop;
    pSample->GetTime(&tStart, &tStop);

	AM_MEDIA_TYPE* pMediaType;
	pSample->GetMediaType(&pMediaType);

	REFERENCE_TIME avgTimePerFrame = 0;
	
	if(pMediaType != NULL)
	{
		if (pMediaType->formattype == FORMAT_VideoInfo)
		{
			if (pMediaType->cbFormat >= sizeof(VIDEOINFOHEADER))
			{
				VIDEOINFOHEADER *pVih = (VIDEOINFOHEADER*)m_mt.pbFormat;
				 avgTimePerFrame = pVih->AvgTimePerFrame;
			}
		}
	}

    DbgLog((LOG_TRACE, 1, TEXT("tStart(%s), tStop(%s), Diff(%d ms), Bytes(%d)"),
           (LPCTSTR) CDisp(tStart),
           (LPCTSTR) CDisp(tStop),
           (LONG)((tStart - m_tLast) / 10000),
           pSample->GetActualDataLength()));

    //m_tLast = tStart;

	// Copy Sample size to a global size.
	lGActualDataLength = pSample->GetActualDataLength();
	//lGActualDataLength = lGActualDataLength -4; // Mpeg4 

	// Copy the data to a global buffer
	HRESULT hr = pSample->GetPointer(&pbGReadBuffer);
    if (FAILED(hr)) {
        return hr;
    }
	
//	pbGReadBuffer = pbGReadBuffer + 4; // Mpeg4 
    fWatchVariable = 1; // Signalinf TaskSchudler for a new frame.
	this->doGetNextFrame();	
	//this->deliverFrame();
	
	gettimeofday(&fPresentationTime,NULL);
	printf("\nRec:-sec=%d,msec=%d",fPresentationTime.tv_sec,fPresentationTime.tv_usec);
	
    //WriteStringInfo(pSample);  from dump filter writing to file

	printf("\nms:%d., size=%d",MediaSampleCounter++,lGActualDataLength);

	return S_OK;
}


//
// WriteStringInfo
//
// Write to the file as text form
//
HRESULT CRTSPInputPin::WriteStringInfo(IMediaSample *pSample)
{
    CheckPointer(pSample,E_POINTER);

    TCHAR TempString[256],FileString[256];
    PBYTE pbData;

    // Retrieve the time stamps from this sample

    REFERENCE_TIME tStart, tStop;
    pSample->GetTime(&tStart, &tStop);
    m_tLast = tStart;

    // Write the sample time stamps out

    HRESULT hr = StringCchPrintf(FileString,256,TEXT("\r\nRenderer received sample (%dms)\0"),timeGetTime());
    m_pRTSP->WriteString(FileString);

    hr = StringCchPrintf(FileString,256,TEXT("   Start time (%s)\0"),(LPCTSTR)CDisp(tStart));
    m_pRTSP->WriteString(FileString);

    hr = StringCchPrintf(FileString,256,TEXT("   End time (%s)\0"),(LPCTSTR)CDisp(tStop));
    m_pRTSP->WriteString(FileString);

    // Display the media times for this sample

    hr = pSample->GetMediaTime(&tStart, &tStop);
    if (hr == NOERROR) 
    {
        hr = StringCchPrintf(FileString,256,TEXT("   Start media time (%s)\0"),(LPCTSTR)CDisp(tStart));
        m_pRTSP->WriteString(FileString);
        hr = StringCchPrintf(FileString,256,TEXT("   End media time (%s)\0"),(LPCTSTR)CDisp(tStop));
        m_pRTSP->WriteString(FileString);
    }

    // Is this a sync point sample

    hr = pSample->IsSyncPoint();
    hr = StringCchPrintf(FileString,256,TEXT("   Sync point (%d)\0"),(hr == S_OK));
    m_pRTSP->WriteString(FileString);

    // Is this a preroll sample

    hr = pSample->IsPreroll();
    hr = StringCchPrintf(FileString,256,TEXT("   Preroll (%d)\0"),(hr == S_OK));
    m_pRTSP->WriteString(FileString);

    // Is this a discontinuity sample

    hr = pSample->IsDiscontinuity();
    hr = StringCchPrintf(FileString,256,TEXT("   Discontinuity (%d)\0"),(hr == S_OK));
    m_pRTSP->WriteString(FileString);

    // Write the actual data length

    LONG DataLength = pSample->GetActualDataLength();
    hr = StringCchPrintf(FileString,256,TEXT("   Actual data length (%d)\0"),DataLength);
    m_pRTSP->WriteString(FileString);

    // Does the sample have a type change aboard

    AM_MEDIA_TYPE *pMediaType;
    pSample->GetMediaType(&pMediaType);
    hr = StringCchPrintf(FileString,256,TEXT("   Type changed (%d)\0"),
        (pMediaType ? TRUE : FALSE));

    m_pRTSP->WriteString(FileString);
    DeleteMediaType(pMediaType);

    // Copy the data to the file

    hr = pSample->GetPointer(&pbData);
    if (FAILED(hr)) {
        return hr;
    }

    // Write each complete line out in BYTES_PER_LINES groups

    for (int Loop = 0;Loop < (DataLength / BYTES_PER_LINE);Loop++) 
    {
        hr = StringCchPrintf(FileString,256,FIRST_HALF_LINE,
                 pbData[0],pbData[1],pbData[2],
                 pbData[3],pbData[4],pbData[5],pbData[6],
                 pbData[7],pbData[8],pbData[9]);

        hr = StringCchPrintf(TempString,256, SECOND_HALF_LINE,
                 pbData[10],pbData[11],pbData[12],
                 pbData[13],pbData[14],pbData[15],pbData[16],
                 pbData[17],pbData[18],pbData[19]);

        hr = StringCchCat(FileString,256,TempString);
        m_pRTSP->WriteString(FileString);
        pbData += BYTES_PER_LINE;
    }

    // Write the last few bytes out afterwards

    hr = StringCchPrintf(FileString,256,TEXT("   \0"));
    for (int Loop = 0;Loop < (DataLength % BYTES_PER_LINE);Loop++) 
    {
        hr = StringCchPrintf(FileString,256,TEXT("%x \0"),pbData[Loop]);
        hr = StringCchCat(FileString,256,TempString);
    }

    m_pRTSP->WriteString(FileString);
    return NOERROR;
}


//
// EndOfStream
//
STDMETHODIMP CRTSPInputPin::EndOfStream(void)
{
    CAutoLock lock(m_pReceiveLock);
    return CRenderedInputPin::EndOfStream();

} // EndOfStream


//
// NewSegment
//
// Called when we are seeked
//
STDMETHODIMP CRTSPInputPin::NewSegment(REFERENCE_TIME tStart,
                                       REFERENCE_TIME tStop,
                                       double dRate)
{
    m_tLast = 0;
    return S_OK;

} // NewSegment


//
//  CRTSP class
//
CRTSP::CRTSP(LPUNKNOWN pUnk, HRESULT *phr) :
    CUnknown(NAME("CRTSP"), pUnk),
    m_pFilter(NULL),
    m_pPin(NULL),
    m_pPosition(NULL),
    m_hFile(INVALID_HANDLE_VALUE),
    m_pFileName(0),
    m_fWriteError(0)
{
    ASSERT(phr);
    
    m_pFilter = new CRTSPFilter(this, GetOwner(), &m_Lock, phr);
    if (m_pFilter == NULL) {
        if (phr)
            *phr = E_OUTOFMEMORY;
        return;
    }

    //m_pPin = new CRTSPInputPin(this,GetOwner(),
    //                           m_pFilter,
    //                           &m_Lock,
    //                           &m_ReceiveLock,
    //                           phr);

	m_pPin = CRTSPInputPin::createNew(*env,this,GetOwner(),
                               m_pFilter,
                               &m_Lock,
                               &m_ReceiveLock,
                               phr);
    if (m_pPin == NULL) {
        if (phr)
            *phr = E_OUTOFMEMORY;
        return;
    }
	mpeg1or2LiveSource = m_pPin;
}


//
// SetFileName
//
// Implemented for IFileSinkFilter support
//
STDMETHODIMP CRTSP::SetFileName(LPCOLESTR pszFileName,const AM_MEDIA_TYPE *pmt)
{
    // Is this a valid filename supplied

    CheckPointer(pszFileName,E_POINTER);
    if(wcslen(pszFileName) > MAX_PATH)
        return ERROR_FILENAME_EXCED_RANGE;

    // Take a copy of the filename

    size_t len = 1+lstrlenW(pszFileName);
    m_pFileName = new WCHAR[len];
    if (m_pFileName == 0)
        return E_OUTOFMEMORY;

    HRESULT hr = StringCchCopyW(m_pFileName, len, pszFileName);

    // Clear the global 'write error' flag that would be set
    // if we had encountered a problem writing the previous RTSP file.
    // (eg. running out of disk space).
    m_fWriteError = FALSE;

    // Create the file then close it

    hr = OpenFile();
    CloseFile();

    return hr;

} // SetFileName


//
// GetCurFile
//
// Implemented for IFileSinkFilter support
//
STDMETHODIMP CRTSP::GetCurFile(LPOLESTR * ppszFileName,AM_MEDIA_TYPE *pmt)
{
    CheckPointer(ppszFileName, E_POINTER);
    *ppszFileName = NULL;

    if (m_pFileName != NULL) 
    {
        size_t len = 1+lstrlenW(m_pFileName);
        *ppszFileName = (LPOLESTR)
        QzTaskMemAlloc(sizeof(WCHAR) * (len));

        if (*ppszFileName != NULL) 
        {
            HRESULT hr = StringCchCopyW(*ppszFileName, len, m_pFileName);
        }
    }

    if(pmt) 
    {
        ZeroMemory(pmt, sizeof(*pmt));
        pmt->majortype = MEDIATYPE_NULL;
        pmt->subtype = MEDIASUBTYPE_NULL;
    }

    return S_OK;

} // GetCurFile


// Destructor

CRTSP::~CRTSP()
{
    CloseFile();

    delete m_pPin;
    delete m_pFilter;
    delete m_pPosition;
    delete m_pFileName;
}


//
// CreateInstance
//
// Provide the way for COM to create a RTSP filter
//
CUnknown * WINAPI CRTSP::CreateInstance(LPUNKNOWN punk, HRESULT *phr)
{
    ASSERT(phr);
    
	
	 // Begin by setting up our usage environment:
	TaskScheduler* scheduler = BasicTaskScheduler::createNew();
    env = BasicUsageEnvironment::createNew(*scheduler);
	
    CRTSP *pNewObject = new CRTSP(punk, phr);
    if (pNewObject == NULL) {
        if (phr)
            *phr = E_OUTOFMEMORY;
    }
	
    return pNewObject;

} // CreateInstance


//
// NonDelegatingQueryInterface
//
// Override this to say what interfaces we support where
//
STDMETHODIMP CRTSP::NonDelegatingQueryInterface(REFIID riid, void ** ppv)
{
    CheckPointer(ppv,E_POINTER);
    CAutoLock lock(&m_Lock);

    // Do we have this interface

    if (riid == IID_IFileSinkFilter) {
        return GetInterface((IFileSinkFilter *) this, ppv);
    } 
    else if (riid == IID_IBaseFilter || riid == IID_IMediaFilter || riid == IID_IPersist) {
        return m_pFilter->NonDelegatingQueryInterface(riid, ppv);
    } 
    else if (riid == IID_IMediaPosition || riid == IID_IMediaSeeking) {
        if (m_pPosition == NULL) 
        {

            HRESULT hr = S_OK;
            m_pPosition = new CPosPassThru(NAME("RTSP Pass Through"),
                                           (IUnknown *) GetOwner(),
                                           (HRESULT *) &hr, m_pPin);
            if (m_pPosition == NULL) 
                return E_OUTOFMEMORY;

            if (FAILED(hr)) 
            {
                delete m_pPosition;
                m_pPosition = NULL;
                return hr;
            }
        }

        return m_pPosition->NonDelegatingQueryInterface(riid, ppv);
    } 

    return CUnknown::NonDelegatingQueryInterface(riid, ppv);

} // NonDelegatingQueryInterface


//
// OpenFile
//
// Opens the file ready for RTSPing
//
HRESULT CRTSP::OpenFile()
{
    TCHAR *pFileName = NULL;

    // Is the file already opened
    if (m_hFile != INVALID_HANDLE_VALUE) {
        return NOERROR;
    }

    // Has a filename been set yet
    if (m_pFileName == NULL) {
        return ERROR_INVALID_NAME;
    }

    // Convert the UNICODE filename if necessary

#if defined(WIN32) && !defined(UNICODE)
    char convert[MAX_PATH];

    if(!WideCharToMultiByte(CP_ACP,0,m_pFileName,-1,convert,MAX_PATH,0,0))
        return ERROR_INVALID_NAME;

    pFileName = convert;
#else
    pFileName = m_pFileName;
#endif

    // Try to open the file

    m_hFile = CreateFile((LPCTSTR) pFileName,   // The filename
                         GENERIC_WRITE,         // File access
                         FILE_SHARE_READ,       // Share access
                         NULL,                  // Security
                         CREATE_ALWAYS,         // Open flags
                         (DWORD) 0,             // More flags
                         NULL);                 // Template

    if (m_hFile == INVALID_HANDLE_VALUE) 
    {
        DWORD dwErr = GetLastError();
        return HRESULT_FROM_WIN32(dwErr);
    }

    return S_OK;

} // Open


//
// CloseFile
//
// Closes any RTSP file we have opened
//
HRESULT CRTSP::CloseFile()
{
    // Must lock this section to prevent problems related to
    // closing the file while still receiving data in Receive()
    CAutoLock lock(&m_Lock);

    if (m_hFile == INVALID_HANDLE_VALUE) {
        return NOERROR;
    }

    CloseHandle(m_hFile);
    m_hFile = INVALID_HANDLE_VALUE; // Invalidate the file 

    return NOERROR;

} // Open


//
// Write
//
// Write raw data to the file
//
HRESULT CRTSP::Write(PBYTE pbData, LONG lDataLength)
{
    DWORD dwWritten;

    // If the file has already been closed, don't continue
    if (m_hFile == INVALID_HANDLE_VALUE) {
        return S_FALSE;
    }

    if (!WriteFile(m_hFile, (PVOID)pbData, (DWORD)lDataLength,
                   &dwWritten, NULL)) 
    {
        return (HandleWriteFailure());
    }

    return S_OK;
}


HRESULT CRTSP::HandleWriteFailure(void)
{
    DWORD dwErr = GetLastError();

    if (dwErr == ERROR_DISK_FULL)
    {
        // Close the RTSP file and stop the filter, 
        // which will prevent further write attempts
        m_pFilter->Stop();

        // Set a global flag to prevent accidental deletion of the RTSP file
        m_fWriteError = TRUE;

        // Display a message box to inform the developer of the write failure
        TCHAR szMsg[MAX_PATH + 80];
        HRESULT hr = StringCchPrintf(szMsg, MAX_PATH + 80, TEXT("The disk containing RTSP file has run out of space, ")
                  TEXT("so the RTSP filter has been stopped.\r\n\r\n")
                  TEXT("You must set a new RTSP file name or restart the graph ")
                  TEXT("to clear this filter error."));
        MessageBox(NULL, szMsg, TEXT("RTSP Filter failure"), MB_ICONEXCLAMATION);
    }

    return HRESULT_FROM_WIN32(dwErr);
}

//
// WriteString
//
// Writes the given string into the file
//
void CRTSP::WriteString(TCHAR *pString)
{
    ASSERT(pString);
    
    // If the file has already been closed, don't continue
    if (m_hFile == INVALID_HANDLE_VALUE) {
        return;
    }

    BOOL bSuccess;
    DWORD dwWritten = 0;
    DWORD dwToWrite = lstrlen(pString);

    // Write the requested data to the RTSP file
    bSuccess = WriteFile((HANDLE) m_hFile,
                         (PVOID) pString, (DWORD) dwToWrite,
                         &dwWritten, NULL);

    if (bSuccess == TRUE)
    {
        // Append a carriage-return and newline to the file
        const TCHAR *pEndOfLine = TEXT("\r\n\0");
        dwWritten = 0;
        dwToWrite = lstrlen(pEndOfLine);

        bSuccess = WriteFile((HANDLE) m_hFile,
                             (PVOID) pEndOfLine, (DWORD) dwToWrite,
                             &dwWritten, NULL);
    }

    // If either of the writes failed, stop receiving data
    if (!bSuccess || (dwWritten < dwToWrite))
    {
        HandleWriteFailure();
    }

} // WriteString


////////////////////////////////////////////////////////////////////////
//
// Exported entry points for registration and unregistration 
// (in this case they only call through to default implementations).
//
////////////////////////////////////////////////////////////////////////

//
// DllRegisterSever
//
// Handle the registration of this filter
//
STDAPI DllRegisterServer()
{
    return AMovieDllRegisterServer2( TRUE );

} // DllRegisterServer


//
// DllUnregisterServer
//
STDAPI DllUnregisterServer()
{
    return AMovieDllRegisterServer2( FALSE );

} // DllUnregisterServer


//
// DllEntryPoint
//
extern "C" BOOL WINAPI DllEntryPoint(HINSTANCE, ULONG, LPVOID);

BOOL APIENTRY DllMain(HANDLE hModule, 
                      DWORD  dwReason, 
                      LPVOID lpReserved)
{
	return DllEntryPoint((HINSTANCE)(hModule), dwReason, lpReserved);
}

void InitRTSP(void *arglist)
//unsigned __stdcall InitRTSP(void *)
{
 //// Begin by setting up our usage environment:
 // TaskScheduler is global created before filter
 // TaskScheduler* scheduler = BasicTaskScheduler::createNew();
 // env = BasicUsageEnvironment::createNew(*scheduler);

  // Create 'groupsocks' for RTP and RTCP:
  char const* destinationAddressStr
#ifdef USE_SSM
    = "232.255.42.42";
#else
    = "239.255.42.42";
  // Note: This is a multicast address.  If you wish to stream using
  // unicast instead, then replace this string with the unicast address
  // of the (single) destination.  (You may also need to make a similar
  // change to the receiver program.)
#endif
  const unsigned short rtpPortNum = 8888;
  const unsigned short rtcpPortNum = rtpPortNum+1;
  const unsigned char ttl = 7; // low, in case routers don't admin scope

  struct in_addr destinationAddress;
  destinationAddress.s_addr = our_inet_addr(destinationAddressStr);
  const Port rtpPort(rtpPortNum);
  const Port rtcpPort(rtcpPortNum);

  Groupsock rtpGroupsock(*env, destinationAddress, rtpPort, ttl);
  Groupsock rtcpGroupsock(*env, destinationAddress, rtcpPort, ttl);
#ifdef USE_SSM
  rtpGroupsock.multicastSendOnly();
  rtcpGroupsock.multicastSendOnly();
#endif

  // Create a 'MPEG Video RTP' sink from the RTP 'groupsock':
  videoSink = MPEG1or2VideoRTPSink::createNew(*env, &rtpGroupsock);

  // Create (and start) a 'RTCP instance' for this RTP sink:
  const unsigned estimatedSessionBandwidth = 4500; // in kbps; for RTCP b/w share
  const unsigned maxCNAMElen = 100;
  unsigned char CNAME[maxCNAMElen+1];
  gethostname((char*)CNAME, maxCNAMElen);
  CNAME[maxCNAMElen] = '\0'; // just in case
#ifdef IMPLEMENT_RTSP_SERVER
  RTCPInstance* rtcp =
#endif
    RTCPInstance::createNew(*env, &rtcpGroupsock,
			      estimatedSessionBandwidth, CNAME,
			      videoSink, NULL /* we're a server */, isSSM);
  // Note: This starts RTCP running automatically

#ifdef IMPLEMENT_RTSP_SERVER
  RTSPServer* rtspServer = RTSPServer::createNew(*env);
  // Note that this (attempts to) start a server on the default RTSP server
  // port: 554.  To use a different port number, add it as an extra
  // (optional) parameter to the "RTSPServer::createNew()" call above.
  if (rtspServer == NULL) {
    *env << "Failed to create RTSP server: " << env->getResultMsg() << "\n";
    exit(1);
  }

  char const* inputFileName = "MyInfoFileName";

  ServerMediaSession* sms
    = ServerMediaSession::createNew(*env, "testStream", inputFileName, "Session streamed by \"testMPEG1or2VideoStreamer\"",isSSM);
  sms->addSubsession(PassiveServerMediaSubsession::createNew(*videoSink, rtcp));
  rtspServer->addServerMediaSession(sms);
  
  char* url = rtspServer->rtspURL(sms);
  *env << "Play this stream using the URL \"" << url << "\"\n";
  delete[] url;
#endif

  //// Finally, start the streaming:
  *env << "Beginning streaming...\n";
  play();

 
  while(true)
	{
		fWatchVariable = 0;
		printf("\nfWatchVariable=%d,",fWatchVariable);				
		dummyTask(NULL);
		env->taskScheduler().doEventLoop(&fWatchVariable);
		printf("\nIn While fWatchVariable=%d,",fWatchVariable);					
	}


// return 0; // only to prevent compiler warning
}

void afterPlaying(void* /*clientData*/) {
  *env << "...done Live Streaming from file\n";

  Medium::close(videoSource);
#ifdef SOURCE_IS_PROGRAM_STREAM
  Medium::close(mpegDemux);
#endif
  // Note that this also closes the input file that this source read from.

  play();
}

void play() {
  // Open the input file as a 'byte-stream file source':
 /* ByteStreamFileSource* fileSource = ByteStreamFileSource::createNew(*env, inputFileName);
  if (fileSource == NULL) {
    *env << "Unable to open file \"" << inputFileName
	 << "\" as a byte-stream file source\n";
    exit(1);
  }*/

  FramedSource* videoES;
#ifdef SOURCE_IS_PROGRAM_STREAM
  // We must demultiplex a Video Elementary Stream from the input source:
  mpegDemux = MPEG1or2Demux::createNew(*env, mpeg1or2LiveSource);
  videoES = mpegDemux->newVideoStream();
#else
  // The input source is assumed to already be a Video Elementary Stream:
  videoES = mpeg1or2LiveSource; // Get from the Global CRTSPInputPin type FramedSource
#endif

  // Create a framer for the Video Elementary Stream:
  videoSource
    = MPEG1or2VideoStreamFramer::createNew(*env, videoES, iFramesOnly);

  // Finally, start playing:
  *env << "Beginning to Live stream...\n";
  //videoSink->startPlaying(*videoSource, afterPlaying, videoSink);
  videoSink->startPlaying(*videoSource, afterPlaying, videoSink);

}